/*
	This app scans for ps2.img files that would ultimately cause
	the quantizer to exceed a specified memory limit
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "PaletteGen.h"

unsigned int gMemUsage = 0;
unsigned int gNumCubes = 0;

enum TextureFormat
{
	v32_BIT,
	v24_BIT,
	v16_BIT,
	v8_BIT,
	v4_BIT,		
	v8_BIT_GRAY,
	v4_BIT_GRAY,
	v2_BIT,
	v1_BIT,
	v2_BIT_GRAY,
	v1_BIT_GRAY,
	vUNKNOWN,
};

enum PSPixelMode
{
	vPSMCT32 = 0,
	vPSMCT24 = 1,
	vPSMCT16 = 2,
	vPSMT8   = 19,
	vPSMT4   = 20,
};

TextureFormat ConvertPSDepth(PSPixelMode mode)
{
	switch(mode)
	{
	case vPSMCT32:
		return v32_BIT;

	case vPSMCT24:
		return v24_BIT;

	case vPSMCT16:
		return v16_BIT;

	case vPSMT8:
		return v8_BIT;

	case vPSMT4:
		return v4_BIT;

	default:
		return (TextureFormat)0;
	}
}

// Returns the number of bytes per entry for a given bit depth
float GetBytesPerDepth(TextureFormat depth)
{
	switch(depth)
	{
	case v2_BIT:
		return 0.25f;

	case v4_BIT:
		return 0.5f;

	case v8_BIT:
		return 1;

	case v16_BIT:
		return 2;

	case v24_BIT:
		return 3;

	case v32_BIT:
		return 4;

	default:
		return 0;		// Unsupported depth
	}
}

int GetDecimalDepth(TextureFormat format)
{
	switch(format)
	{
	case v32_BIT:
		return 32;

	case v24_BIT:
		return 24;
		
	case v16_BIT:
		return 16;

	case v8_BIT:
		return 8;

	case v4_BIT:
		return 4;

	case v8_BIT_GRAY:
		return 8;

	case v4_BIT_GRAY:
		return 4;

	case v2_BIT:
		return 2;

	case v1_BIT:
		return 1;

	case v2_BIT_GRAY:
		return 2;

	case v1_BIT_GRAY:
		return 1;

	default:
		return 0;
	}
}

int GetNumPaletteEntries(TextureFormat bppMode)
{
	// Only 8 bit and 4 bit modes have palettes
	switch(bppMode)
	{
	case v8_BIT:
		return 1 << 8;

	case v4_BIT:
		return 1 << 4;

	default:
		return 0;
	}
}

struct ImgInfo
{
	unsigned char* image;		// Pointer to actual image data
	unsigned char* palette;
	unsigned int   width;
	unsigned int   height;
	int            pixelDepth;
	int            paletteDepth;
	int            numPalColors;
};

ImgInfo LoadPS2File(char* filename)
{
	ImgInfo imgInfo;

	unsigned char* image    = NULL;
	unsigned char* pPalette = NULL;

	// Load in the file
	FILE* fp = fopen(filename, "rb");

	if (!fp)
	{
		printf("Failed to open '%s'.", filename);
		imgInfo.image = NULL;
		return imgInfo;
	}

/*
Offset		Size (Bytes)	Field
---------------------------------------------------------------
0x00			4			Version (currently 1)
0x04			4			Basename of file in CRC
format
0x08			4			PS2 TW (log2 Width)
0x0c			4			PS2 TH (log2 Height)
0x10			4			PS2 PSM (Bit depth)
0x14			4			PS2 CPSM (CLUT bit
depth)
0x18			4			PS2 MXL (Number of
mipmaps)
0x1c			2			Original image width
0x1e			2			Original image height
0x20			Upto 1K		CLUT data
0x20+CLUT_size	W*H*depth		Texture mipmaps
*/
	int version;
	unsigned long crc;
	int log2Width, log2Height;
	TextureFormat bitDepth;
	TextureFormat clutDepth;
	int nMips;
	short width, height;

	fread(&version, sizeof(int), 1, fp);
	fread(&crc, sizeof(unsigned long), 1, fp);
	fread(&log2Width, sizeof(int), 1, fp);
	fread(&log2Height, sizeof(int), 1, fp);
	fread(&bitDepth, sizeof(TextureFormat), 1, fp);
	fread(&clutDepth, sizeof(TextureFormat), 1, fp);
	fread(&nMips, sizeof(int), 1, fp);
	fread(&width, sizeof(short), 1, fp);
	fread(&height, sizeof(short), 1, fp);

	unsigned int imgWidth, imgHeight;
	int paletteDepth, pixelDepth;

	bitDepth = ConvertPSDepth((PSPixelMode)bitDepth);

	pixelDepth   = GetDecimalDepth(bitDepth);
	paletteDepth = GetDecimalDepth(clutDepth);
	imgWidth     = width;
	imgHeight    = height;

	imgInfo.pixelDepth   = pixelDepth;
	imgInfo.paletteDepth = paletteDepth;
	imgInfo.width        = width;
	imgInfo.height       = height;
	//imgInfo.numPalColors = nPaletteEntries;

	if (pixelDepth == 2)
	{
		//MessageBox(hwndOwner, "2-bit images cannot currently be displayed in this viewer.", "Image can't be displayed", MB_ICONWARNING|MB_OK);
		printf("2-bit images are not supported\n");
		fclose(fp);
		imgInfo.image = NULL;
		return imgInfo;
	}

	// Check for gray scale images
	if (bitDepth  == v8_BIT_GRAY ||
		bitDepth  == v4_BIT_GRAY ||
		bitDepth  == v2_BIT_GRAY ||
		bitDepth  == v1_BIT_GRAY ||
		clutDepth == v8_BIT_GRAY ||
		clutDepth == v4_BIT_GRAY ||
		clutDepth == v2_BIT_GRAY ||
		clutDepth == v1_BIT_GRAY)
	{
		printf("Gray scale images are not currently supported.\n");
		fclose(fp);
		imgInfo.image = NULL;
		return imgInfo;
	}

	// Load in the CLUT
	int nCLUTEntries = GetNumPaletteEntries(bitDepth);
	int nPaletteEntries = nCLUTEntries;

	if (pPalette)
	{
		free(pPalette);
		pPalette = NULL;
	}

	if (nCLUTEntries > 0)
	{
		// Ensure that the palette is in a valid format
		if (clutDepth != v16_BIT &&
			clutDepth != v24_BIT &&
			clutDepth != v32_BIT)
		{
			printf("Palettized image uses a non-supported palette depth.  Only 16/24/32 bitdepths are supported.\n");
			fclose(fp);
			imgInfo.image = NULL;
			return imgInfo;
		}

		int bytesPerEntry = (int)GetBytesPerDepth(clutDepth);
		pPalette = (unsigned char*)malloc(bytesPerEntry * nCLUTEntries);
		
		fread(pPalette, bytesPerEntry, nCLUTEntries, fp);		
	}

	// Read in the image (only the first mip gets read)
	if (image)
	{
		free(image);
		image = NULL;
	}

	image = (unsigned char*)malloc((int)(GetBytesPerDepth(bitDepth) * (float)width * (float)height));
	fread(image, (int)(GetBytesPerDepth(bitDepth) * (float)width * (float)height), 1, fp);

	imgInfo.image   = image;
	imgInfo.palette = pPalette;

	fclose(fp);
	return imgInfo;
}

#define MAINLOOP_MEMMARKER  16   // from Engine.cpp ???

struct Entry
{
	unsigned long val;
	char          name[40];
};

int compare(const void* v1, const void* v2)
{
	Entry* val1 = (Entry*)v1;
	Entry* val2 = (Entry*)v2;

	if (val1->val == val2->val)
		return 0;

	if (val1->val < val2->val)
		return -1;

	return 1;
}

int main(int argc, char* args[])
{
	int lastMemID = 0;

	char* image = NULL;

	if (argc < 2)
	{
		printf("USAGE: QuantCheck [name] [-b]\n");
		printf("\t-b batch mode\n");
		return (int)false;
	}

	//Mem::Manager::sSetUp();
	//Mem::Manager::sHandle().InitOtherHeaps();
	//Mem::Manager::sHandle().BottomUpHeap()->PushContext();
	//Mem::Manager::sHandle().PushMemoryMarker(MAINLOOP_MEMMARKER);

	// Just test a single file
	ImgInfo imgInfo = LoadPS2File(args[1]);

	if (!imgInfo.image)
	{
		printf("File '%s' failed to load.\n", args[1]);
		return (int)false;
	}

	// Quick and dirty mem usage check
	gMemUsage = 0;

	// We need to convert from palettized format (assuming 8-bit [256 colors]) to non palettized
	int size = imgInfo.width * imgInfo.height;
	unsigned int* newImg = (unsigned int*)malloc(size * 4);
	int pos = 0;

	for(int i = 0; i < size; i++)
	{
		newImg[pos] = ((unsigned int*)imgInfo.palette)[imgInfo.image[i]];
		pos++;
	}	

	// Attempt quantization and discover memory usage
	Nx::PalQuant pq;
	Nx::PalQuantInit(&pq);

	Image::RGBA palette[256];

	Nx::GeneratePalette( palette, (uint8*)newImg, imgInfo.width, imgInfo.height, imgInfo.paletteDepth, 256 );

	//printf("%s uses %i bytes of memory.", args[1], gMemUsage);	
	printf("%i\n", gMemUsage);
	free(newImg);

	exit(gMemUsage);
	return gMemUsage;
}
